#!/data/oiropt/bin/ksh 
# $Id: skyproc.ksh,v 1.9 2006/03/24 17:58:05 wyatt Exp $
#
# Generate starbase tables of:
#
#   1) All the keywords in all FITS files     --> ${dirname}_keyval.db
#   2) All Object files                       --> ${dirname}_objlist.db
#   3) All Skyfiles & Certification values    --> ${dirname}_skylist.db
#   4) Objects and distances to each Sky file --> ${dirname}_skydist.dd
#
# Number 4 uses either
#   (a) keywords FIBER_X and FIBER_Y to compute the distance
#       between an object and each of the sky files in turn.
#   (b) APERTURE to get the closest apertures containing skys to
#       each of the objects.
#

export DEFAULT_DIR=${iraf}../iraf.pkg/hectospec/
export PATH=/data/oiropt/starbase:/data/oiropt/bin:${PATH}

ulimit -s unlimited

typeset -F FiberRadiusLimit=306.0 # limit of non-parked fibers

typeset -F mean stddev
typeset -F rmsstd skyfactor skyweight
typeset -l skysort

verbose=1
destdir="../skysub/"
files=\*.fits
mpjobs=0
nskys=6
skysort="skydist"     # also 'aperture' is possible 
skymethod="SKYSUB"    # also 'SKYFIT' is now available using svdfit
logfile="STDOUT"
fiberdist=0.
skysort="skydist"
skysame=1
skycert=2.9
skystdlim=0.2
emline="OH8399"
usearea=0        # use area if 1, line height otherwise in sky scaling
skyweight="0.0"  # use sky ratio if 0, otherwise this factor (i.e. 1.0)

uparmdir=${uparm:-$UPARM}

# these should be skyproc params
ab_contam="fabtemp97,habtemp90,m31_a_temp,m31_f_temp,m31_k_temp"
em_contam="femtemp97"
#em_contam="hemtemp0.0"

while getopts "a:c:d:e:f:i:j:k:l:m:n:r:s:u:v:w:hx" fval ; do
   case $fval in
    a) if   [[ $OPTARG == yes ]]; then usearea=1
       elif [[ $OPTARG == no  ]]; then usearea=0
       elif [[ $OPTARG -ne 0  ]]; then usearea=1
       elif [[ $OPTARG -eq 0  ]]; then usearea=0
       fi;;
    c) if   [[ $OPTARG == yes ]]; then skysame=1
       elif [[ $OPTARG == no  ]]; then skysame=0
       elif [[ $OPTARG -ne 0  ]]; then skysame=1
       elif [[ $OPTARG -eq 0  ]]; then skysame=0
       fi;;
    d) destdir=$OPTARG   ;;
    e) emline=$OPTARG    ;;
    f) fiberdist=$OPTARG ;;
    i) files="$OPTARG"   ;;
    j) mpjobs=$OPTARG    ;;
    l) logfile=$OPTARG   ;;
    k) skysort=$OPTARG   ;;
    m) skymethod=$(echo $OPTARG | tr '[:lower:]' '[:upper:]');;
    n) nskys=$OPTARG     ;;
    r) skycert=$OPTARG   ;;
    s) skystdlim=$OPTARG ;;
    u) uparmdir=$OPTARG  ;;
    v) verbose=$OPTARG   ;;
    w) skyweight=$OPTARG ;;
   h|x) print  "Usage: $(basename $0) [options]"
        printf "  -i %s\tinput file list\n" "$files"
        printf "  -d %s\tdestination directory for output files\n" $destdir
        printf "  -n %d\t\tMax number of skys to coadd\n" $nskys
        printf "  -f %.2f\tMax fiber distance to sky\n" $fiberdist
        printf "  -r %.2f\tMax R value for certified sky\n" $skycert
        printf "  -s %.2f\tMax std of difference from avg sky\n" $skystdlim
        printf "  -c %d\t\tUse skys only from same chip: %s\n" $skysame \
                "$(if [[ $skysame -gt 0 ]]; then echo yes; else echo no;fi)"
        printf "  -e %s\t\tEmission line to scale by: O5577, OH8399\n" "$emline"
        printf "  -a %d\tUse line area in sky scaling instead of height\n" \
                     $usearea
        printf "  -m %s\tSky removal method, skysub or skyfit\n" "$skymethod"
        printf "  -w %.1f\tIf not 0.0, use this weight instead of sky ratio\n" \
                     $skyfactor
        printf "  -k %s\tSky sorting type\n" $skysort
        printf "  -l %s\tName of output summary file\n" $logfile
        printf "  -j %d\t\tMax number of multiprocessor jobs in parallel\n" \
                         $mpjobs
        printf "  -u %s\tName of uparm directory\n" "${uparmdir:-$UPARM}"
        printf "  -v %d\t\tverbose - output summary (on stderr):\n" $verbose 
        printf "\t\t  0=quiet, 1=notes, 2=very verbose, 3=watch out\n"
        printf "  -(h|x)\t\tprint this usage and exit\n"
        printf  "\n"
        exit 1;;
    \?) print -u2 "Unknown option $OPTARG"; exit 1;;
     :) print -u2 "-$OPTARG requires an argument"; exit 1;;
   esac
done

if (( mpjobs > 4096 )); then
    mpjobs = 4096
elif (( mpjobs < 1 )); then
  if [[ $(uname -s) == "Linux" ]]; then
    ncpu=$(grep -i 'physical id' /proc/cpuinfo | uniq | wc -l)
  else
    ncpu=$(psrinfo | wc -l)
  fi
  if ((ncpu > 4)); then mpjobs=$ncpu; else mpjobs=$((ncpu * 2)); fi
fi

# Make sure output directory is there
destdir=$(pwd)/$destdir
if [[ -d $destdir ]] then
    x=0
else
    mkdir -p $destdir
fi

trap exit ERR

#export archive=$irafpkg/saotdc/pkg/archive/
export archive=$irafpkg/rvsao/
export imextn="oif:imh fxf:fits,fit,fts plf:pl qpf:qp stf:hhh,??h"

skyfit=0
if [[ $skymethod == SKYFIT ]]; then
  skyfit=1
  skyweight=0.0
fi

REV=5.0

tmp="/tmp/$(basename $0).$$"/
mkdir -p $tmp

tmpfileA=${tmp}$$A
tmpfileB=${tmp}$$B
tmpskylist=${tmp}$$.skyhdr

procpid=$$

exitval=0

trap 'print -u2 -f "ERROR $? in $0 at line $LINENO\n"; exitval=1; exit' ERR

trap "/bin/rm -rf $tmp" EXIT INT TERM HUP 

# Function to check for IRAF param files, and
# copy parameter file into , looking in all the normal places
# e.g IRAF uparm, $irafpkg/hectospec, 
# if destdir exists, copy it there, else to $UPARM

                  #  example args: 
                  #   routine   parent   package  IRAF uparm  Destination
function getirafparm #imdelete skyproc  hectospec $uparmdir     $destdir
{
  typeset parname=$(basename $1 .par)  # base name, e.g. "imdelete"
  typeset irafpackname         # name in IRAF package, e.g. "skimdelete"
  typeset userparname          # name in IRAF user parm, e.g. "hecskimde"
  typeset scriptname=$2
  typeset prfix1               # 1st 2, last letter of package, e.g. "hec"
  typeset prfix2="${2:0:2}"    # 1st 2 letters of script name, e.g. "sk"
  typeset userparmdir=$4
  typeset destination=$5
  typeset len

  userparmdir=${userparmdir:-$uparm}
  destination=${destination:-$UPARM}
  irafpackname=${prfix2}$parname

# get 1st 2 letters, last letter of package name
  len=$((${#3} - 1))
  if ((len <= 2)); then prfix1=$2; else prfix1="${3:0:2}${3:$len}"; fi

# get 1st 4 letters, last letter of package parameter file name
  len=$((${#irafpackname} - 1))
  if ((len <= 5)); then 
      userparname=${prfix1}${irafpackname}.par  # e.g. "hecskimde.par"
  else 
      userparname="${prfix1}${irafpackname:0:5}${irafpackname:$len:1}.par"
  fi

  if [[ ! (-a $userparmdir/$userparname) ]]; then
    cp $irafpkg/$3/${irafpackname}.par $userparmdir/$userparname
  fi

  cp $userparmdir/$userparname $destination/${parname}.par
}

# Initialize parameter files from hectospec package if not in uparm directory
# and copy to destination directory for use by subroutines
#
# E.g. home$uparm/hecskemso.par  -> $destdir/emsao.par
#  OR
#      $iraf.pkg/hectospec/skemsao.par -> 
#           home$uparm/hecskemso.par and also -> $destdir/emsao.par
#

getirafparm skyproc      ""      hectospec $uparmdir $destdir # hecskyprc.par
getirafparm imarith      skyproc hectospec $uparmdir $destdir # hecskimah.par
getirafparm imcombine    skyproc hectospec $uparmdir $destdir # hecskimce.par
getirafparm imdelete     skyproc hectospec $uparmdir $destdir # hecskimde.par
getirafparm imexpr       skyproc hectospec $uparmdir $destdir # hecskimer.par
getirafparm imrename     skyproc hectospec $uparmdir $destdir # hecskimre.par
getirafparm imstack      skyproc hectospec $uparmdir $destdir # hecskimsk.par
getirafparm imstatistics skyproc hectospec $uparmdir $destdir # hecskimss.par
getirafparm xcsao        skyproc hectospec $uparmdir $destdir # hecskxcso.par
getirafparm emsao        skyproc hectospec $uparmdir $destdir # hecskemso.par
getirafparm contpars     ""      rvsao     $uparmdir $destdir # rvocontps.par

# now make $destdir == the unixiraf UPARM dir

uparmdir=$destdir
UPARM=${uparmdir%/}/
export UPARM
if ((verbose > 0)) then
    echo Parameter files are in $UPARM
fi

( cd $UPARM
  mv imarith.par      imlimarih.par
  mv imcombine.par    imhimcome.par
  mv imdelete.par     imlimdele.par
  mv imexpr.par       imlimexpr.par
  mv imrename.par     imlimrene.par
  mv imstack.par      imlimstak.par
  mv imstatistics.par imlimstas.par
  mv emsao.par	      rvoemsao.par
  mv xcsao.par        rvoxcsao.par
  mv contpars.par     rvocontps.par
  mv skyproc.par      hecskyprc.par
)

trap "/bin/rm -rf $tmp"  INT TERM HUP 
trap "/bin/rm -rf $tmp; \
  [[ -e $destdir/rvoemsao.par ]] && \
     mv $destdir/rvoemsao.par $destdir/hecskemso.par; \
  [[ -e $destdir/rvoxcsao.par ]] && \
     mv $destdir/rvoxcsao.par $destdir/hecskxcso.par; \
  [[ -e $destdir/rvocontps.par ]] && \
     mv $destdir/rvocontps.par $destdir/hecskcons.par " EXIT

# String prefacing output .db files - default is the directory name.
prefix=${prefix:-$(basename $(pwd))_}
if [[ $prefix == "NONE" ]]; then prefix=""; fi

if [[ -z "$logfile" ]] then  logfile="/dev/null"; fi

if [[ $logfile != "-" && $logfile != "STDOUT" ]] then
  exec 1>> $logfile
fi

if (( verbose > 0 )) then
    printf "\n%s rev %s start at $(date)\n\n" $(basename $0) $REV
    printf "  Sky removal method = %s\n" $skymethod
    printf "  FiberRadiusLimit = %.1f\n" $FiberRadiusLimit
    printf "  Sky sorting type = %s\n" $skysort
    printf "  Nskys = %d\n  FiberDist = %.1f\n  SkyCert  <= %.1f\n" \
              $nskys $fiberdist $skycert
    printf "  SkyStdLim = %.2f\n" $skystdlim
    printf "  Skys on same chip = %s\n" \
      $(if [[ $skysame -ge 1 ]]; then echo yes; else echo no; fi)
    if [[ $skymethod == SKYSUB ]]; then
      printf "  Sky line used for scale: %s\n" $emline
      printf "  Sky scaling: scale "
        if((abs(skyweight) > 0.001)); then printf "fixed at %.1f\n" $skyweight
        elif ((usearea > 0));         then printf "by line area\n"
        else                               printf "by line height\n"; fi
    fi
    printf "  DB Files prefix = %s\n" $prefix
    printf "  DestDir = %s\n\n" $destdir
fi

# Generate master table of keywords; works for FITS or IRAF files
files=$(echo "$files" |tr "," " ")
eval gethead -tha IMAGETYP OBJECT APERTURE   \
           FIBER FIBER_X FIBER_Y "$files"   |
  trim   FILENAME=HDRFILE -a  CHIPNUM FIBER_RADIUS                       |
  compute 'if(OBJECT ~ /_-1$/) {OBJECT = "UNUSED"; IMAGETYP = "UNUSED" }
           if(OBJECT ~ /_-2$/) {OBJECT = "BROKEN"; IMAGETYP = "BROKEN" }
           if(OBJECT ~ /_0$/ ) {OBJECT = "SKY"}
           FIBER_RADIUS = sqrt(FIBER_X * FIBER_X + FIBER_Y * FIBER_Y)
           if(APERTURE <= 150) { CHIPNUM = 1 * '$skysame'
             } else { CHIPNUM = 2 * '$skysame' }
  ' |
  sorttable HDRFILE >${prefix}keyval.db

# Pull all [object and not sky] files into a master list
row <${prefix}keyval.db \
   'toupper(IMAGETYP) == "OBJECT" && toupper(OBJECT) !~ /(UNUSED|PARKED|BROKEN)/ && 
        FIBER_RADIUS <= '$FiberRadiusLimit   |
column HDRFILE OBJECT CHIPNUM APERTURE FIBER FIBER_X FIBER_Y    \
>${prefix}objlist.db 

# create a sky file db
row 'FIBER_RADIUS <= '$FiberRadiusLimit' && 
     toupper(IMAGETYP) == "OBJECT" && toupper(OBJECT) ~ "SKY"' <${prefix}keyval.db |
column HDRFILE CHIPNUM APERTURE skyavstd   |
compute 'skyavstd = 0.0'          |
sorttable CHIPNUM HDRFILE >${tmp}skyavstd.db

nsky=$(wctable -l <${tmp}skyavstd.db)
if ((nsky <= 0)); then
mv ${tmp}skyavstd.db .
   echo "No sky spectra in this directory"
   exit -1
fi

column HDRFILE CHIPNUM APERTURE skyavstd <${tmp}skyavstd.db > ${tmp}sky.db

# create the sky file list (not a .db file).
chipval="."   
if [ skysame -gt 0 ]; then chipval="1 2"; fi

# compute RMS of individual skys against the sum of them; test for rejection
for chip in $chipval; do
  rmsstd=0.0
  avout=${tmp}skyav.${chip}.fits
  row 'CHIPNUM ~ /'$chip'/' <${tmp}sky.db >${tmp}chip.db
  skylast=$(wctable -l <${tmp}chip.db)

  for iter in 1 2 3 ; do
    row 'skyavstd <= 3. * '$rmsstd  <${tmp}chip.db | 
    tee ${tmp}chipx.db   |
    headoff HDRFILE  >$tmpskylist
    skys=$(wc -l <$tmpskylist)
 
   ((skys < 1 || iter > 2 && skys == skylast)) && break
   
    skylast=$skys
    # compute master sky average, using 'valid' skys
    /bin/rm -f $avout
    callimages imcombine @$tmpskylist $avout 'rejmask=""' 'combine="average"'\
        'reject="avsigclip"'  'outtype="real"' 'scale="none"' 'zero="none"'  \
        'weight="!EXPTIME"'  mclip=yes  lsigma=3. hsigma=3. sigscale=0.1     \
        'logfile=""'

    # then ratio each against the average to pick
    # up abberant skys to remove from the master sky list
    sed <$tmpskylist -e 's/\.sky_.*\./.ratio./' -e "s,^,$tmp," >$tmpfileA
  
    callimages imarith @$tmpskylist / $avout @$tmpfileA
    callimages imstatistics @$tmpfileA 'fields="stddev"' format=no  |
    fldtotable skyavstd     >$tmpfileB

    pastetable ${tmp}chipx.db $tmpfileB              |
    column -a skyavstd_2=skyavstd -x skyavstd_1      |
    tee ${tmp}chip.db     |
    column  skyavstd      |
    rmstable              |
    headoff               | read rmsstd

    column <${tmp}chip.db skyavstd | 
    maxtable              |
    headoff               | read maxstd

    xargs /bin/rm -rf <$tmpfileA

#print -u2 "Iter ${iter}, Nskys = ${nskys} (${nskylast}), RMS = $rmsstd"
#    [[ $rmsstd -le $((3.0 * skystdlim)) ]] && break
  done

 # redo final calc to get std for all skys, using last avout
  row 'CHIPNUM ~ /'$chip'/' <${tmp}sky.db   |
  headoff HDRFILE >$tmpskylist    
  skys=$(wc -l <$tmpskylist)
  sed <$tmpskylist -e 's/\.sky_.*\./.ratio./' -e "s,^,$tmp," >$tmpfileA

  callimages imarith @$tmpskylist / $avout @$tmpfileA
  callimages imstatistics @$tmpfileA 'fields="mean,stddev"' format=no  |
  while read mean stddev; do 
    printf "%d\t%s\t%s\t%s\n" $skys $mean $stddev $avout
  done 
  xargs /bin/rm -rf <$tmpfileA
done |
fldtotable nskys skyavmean skyavstd avskyfile > ${tmp}skyav.db 

pastetable ${tmp}sky.db ${tmp}skyav.db     |
column HDRFILE APERTURE CHIPNUM nskys skyavmean skyavstd_2=skyavstd avskyfile |
sorttable CHIPNUM HDRFILE >${tmp}skyavstd.db

# generate the sky files each minus the average sky
headoff <${tmp}skyavstd.db HDRFILE >$tmpskylist
headoff <${tmp}skyavstd.db avskyfile >$tmpfileA
sed -e 's/\.sky_.*\./.diff./' -e "s,^,$tmp," <$tmpskylist >$tmpfileB

callimages imarith @$tmpskylist - @$tmpfileA @$tmpfileB

callrvsao xcsao.report_mode=-8\; xcsao.displot=no

# cross-correlate each against ...
( # ... emission templates
  # note: param file is the 'skxcsao' aka hecskxcso.par
 mkdir -p ${tmp}xc1  2>/dev/null
  cp $destdir/rvoxcsao.par ${tmp}xc1
  UPARM=${tmp}xc1/; export UPARM
  callrvsao xcsao @$tmpfileB 'templates="'$em_contam'"' mode=q  |
  column xcr=SkyRem >${tmp}skyxem.db  
) &

( # ... and absorption templates
 mkdir -p ${tmp}xc2  2>/dev/null
  cp $destdir/rvoxcsao.par ${tmp}xc2
  UPARM=${tmp}xc2/; export UPARM
  # note: param file is the 'skxcsao' aka hecskxcso.par
  callrvsao xcsao @$tmpfileB  mode=q 'templates="'$ab_contam'"' |
  column xcr=SkyRab >${tmp}skyxab.db
) &
wait

pastetable ${tmp}skyavstd.db ${tmp}skyxem.db ${tmp}skyxab.db |
column -a SkyR -x avskyfile                                  |
compute 'SkyR = SkyRem; if(SkyRab > SkyRem) { SkyR = SkyRab }' \
 >${prefix}skylist.db

nsky=$(wc -l <${prefix}skylist.db)
if ((nsky <= 2)); then
  echo "No usable sky spectra in this directory"
  exit -1
fi

# Pull join (cprodtable) all certified skyfiles to object list, compute 
# focal plane distance from each object to each sky, also same for aperture 
# number on chip.
jointable -j HDRFILE ${prefix}keyval.db ${prefix}skylist.db  |
column HDRFILE CHIPNUM_1=CHIPNUM APERTURE_1=APERTURE FIBER FIBER_X FIBER_Y    \
            nskys=Nskys SkyR skyavstd=SkyAvStd -x APERTURE_2           |
row '(SkyR ~ /^ *$/ || SkyR <= '$skycert') && SkyAvStd < '$skystdlim >$tmpfileA
cprodtable  ${prefix}objlist.db $tmpfileA                              |
    row 'CHIPNUM_1 == CHIPNUM_2'                                       |
    column HDRFILE_1=HDRFILE HDRFILE_2=Skyfile                   \
          APERTURE_1=APERTURE APERTURE_2=SkyAPERTURE             \
          FIBER_1=FIBER FIBER_X_1=FIBER_X FIBER_Y_1=FIBER_Y      \
          FIBER_2=SkyFIBER FIBER_X_2=SkyX FIBER_Y_2=SkyY         \
          Nskys  SkyDist AperDist SkyR SkyAvStd                        |
    compute 'SkyDist=sprintf("%.1f",                              \
           sqrt( (FIBER_X-SkyX) * (FIBER_X-SkyX) + (FIBER_Y-SkyY) * (FIBER_Y-SkyY)))
           AperDist=abs(APERTURE - SkyAPERTURE)
    ' >${tmp}skydist.db

if [[ $skysort == sky* ]]; then
  # leave sorted by object and then sky distances
    sorttable <${tmp}skydist.db HDRFILE SkyDist:n >${prefix}skydist.db
elif [[ $skysort == aper* ]]; then
  # leave sorted by object and then aperture distances
    sorttable <${tmp}skydist.db HDRFILE AperDist:n >${prefix}skydist.db
elif ((verbose > 0)); then
  printf "\nskyproc: No Skysort specified. (%s)\n" "$skysort"
fi

# just to see the final table
((verbose > 2)) &&
  justify <${prefix}skydist.db                                 \
      HDRFILE,7 Skyfile,7 FIBER,4 FIBER_X,8 FIBER_Y,8          \
      SkyFIBER,6 SkyX,8 SkyY,8 SkyDist,7 SkyR,4 AperDist=ApDist,6 |
  sed 's/	/ /g' 

((verbose > 0)) && {
  printf "\nElapsed time after object & sky tables creation: %.2f seconds\n" \
    $SECONDS
  times
  printf "\n"
}

# run emsao no list of object files to get strength of the selected
# sky line (O5577 or OH8399) line, and add that column to table.

### Used for method = SKYSUB  (skyfit==0)

if ((skyfit == 0)); then
  callrvsao emsao.report_mode=-4\; emsao.displot=no\; emsao.esmooth=2\;        \
     emsao.mincont=10000.\; 'emsao.linedir="rvsao$lib/"'\; contpars.order=10\; \
     'emsao.emsearch="search'$emline'.dat"'\; 'emsao.emlines="'$emline'.dat"'
  
  headoff HDRFILE <${prefix}objlist.db >$tmpfileA
  
  seqno=0
  nsplit=$(( ($(wc -l <$tmpfileA) / mpjobs) + 1))
  split -l $nsplit $tmpfileA ${tmp}objpart.
  
  for objpart in ${tmp}objpart.*; do
    seqno=$((seqno + 1))
    ( 
      xit=0
      tdir=${tmp}objdir.$seqno/
      mkdir -p $tdir 2>/dev/null
      cp $UPARM/rvoemsao.par $UPARM/rvocontps.par $tdir
      UPARM=$tdir
      export UPARM
  
      callrvsao emsao @$objpart  |
      column specid=HDRFILE lht01=objEmiss leqw01=objArea  >${objpart}.db
  
      xit=$?
      [[ $xit != 0 ]] && {
        print -u2 "$(basename $0): Error=$xit in emsao, seq=$seqno \?\n"
      }
      exit $xit
    ) &  # END FORKED SUBSHELL
    
    ((verbose > 1)) &&
      printf " Run emsao on %s, %d files\n" $objpart $(wc -l <$objpart)
  done  
  # wait for the subprocesses to finish
  wait
  
#  mkdir -p /tmp/$LOGNAME 2>/dev/null
#  cp ${tmp}objpart*.db /tmp/$LOGNAME
    
  # Pull the sky tables all together
  if (( $(echo ${tmp}objpart*.db | wc -w) > 1)); then
    uniontable ${tmp}objpart*.db | sorttable HDRFILE >$tmpfileB
  else
    sorttable HDRFILE <${tmp}objpart.aa.db >$tmpfileB
  fi
  
  jointable -j HDRFILE ${prefix}objlist.db $tmpfileB >$tmpfileA
  #cp $tmpfileA ./fileAobjlist.db
  
  if (( $(wctable -l <$tmpfileA) < 1)); then 
     mv $tmpfileB emsao.db
     printf "No rows in jointable?\n"
     exit 1
  fi
  mv $tmpfileA ${prefix}objlist.db
  /bin/rm -rf ${tmp}objpart.* ${tmp}objdir.* $tmpfileB.*
  
  ((verbose > 0)) && {
     printf "\nElapsed time after object emission line fit: $SECONDS seconds\n"
     times
  }

else
  column < ${prefix}objlist.db -a objEmiss objArea >$tmpfileA
  compute ' objEmiss = 0.0;  objArea=0.0 ' <$tmpfileA >${prefix}objlist.db
fi 

#
# Create entries in table for sky summation list for each object
#
seqno=0
waitjob=0
/bin/rm -f ${tmp}nskys.*

for obj in $(headoff <${prefix}objlist.db HDRFILE); do
   seqno=$((seqno + 1))
 (
   tmpA=${tmpfileA}.${seqno}

# Select all skyfiles that are certified and within FIBERDIST of object.
# Note that the test for samechip status (if enabled) has already been done.
   trim <${prefix}skydist.db    |
   table -h '
     HDRFILE == "'$obj'"                             && \
     (SkyR <= '$skycert' || SkyR ~ /^ *$/)           && \
     ('$fiberdist' == 0  || SkyDist <= '$fiberdist') {print Skyfile}
   ' >$tmpA

# Limit to the max allowed number of skys. This chooses the closet ones 
# because the input table was sorted secondarily by SkyDist or AperDist
  if ((nskys > 0)); then
    head -${nskys} $tmpA      |
    sort >${tmpskylist}.$seqno
  else
    sort <$tmpA >${tmpskylist}.$seqno
  fi

# ERROR if no sky files left after selection
  if [[ ! -s ${tmpskylist}.$seqno ]]; then    
    print -u2 -f "\nWARNING:" 
    print -u2 -f " File %s has no sky files within the %.3f mm limit!\n"   \
       $obj $fiberdist
    exitval=1
    continue
  fi
  nskyfiles=$(wc -l <${tmpskylist}.$seqno)
  skylist=$(tr '\012' '+' <${tmpskylist}.$seqno)
  skylist=${skylist%+}
  aplist=$(gethead -a @${tmpskylist}.$seqno aperture | 
            cut -f 2 -d\  | 
            tr '\012' '+')
  if ((verbose > 0)); then
    printf "%s\t%s\t%d\t%s\n" $obj ${skylist%_} $nskyfiles ${aplist%+} \
        >>${tmp}nskys.$seqno
  fi
 ) &
  if ((seqno - waitjob >= mpjobs)); then
      waitjob=$((waitjob + 1))
      wait %${waitjob}
  fi
done 

wait 

cat ${tmp}nskys.*                                |
fldtotable HDRFILE Skyfilelist Nskys Apertures   | 
sorttable HDRFILE >$tmpfileB

jointable -j HDRFILE ${prefix}objlist.db $tmpfileB  >$tmpfileA

mv $tmpfileA ${prefix}objlist.db

/bin/rm -f ${tmp}nskys.*   ${tmpskylist}.*

((verbose > 1)) && {
  printf "\nSet of skys for object files:\n"
  justify <${prefix}objlist.db HDRFILE Nskys Apertures
}


#
# Create sorted, unique set of skyfile lists 
# (with emission line peak value if skyfit==0).
#
skynum=0

headoff Skyfilelist <${prefix}objlist.db | sort -u >${tmp}skyfilelist

seqno=0
nsplit=$(( ($(wc -l <${tmp}skyfilelist) / mpjobs) + 1))
split -l $nsplit ${tmp}skyfilelist ${tmp}skyflist.

for skylist in ${tmp}skyflist.*; do
  seqno=$((seqno + 1))
  skynum=0
  /bin/rm -rf ${skylist}.db

# fork off for multiproc
  (
    xit=0
    tdir=${skylist}.$seqno/
    mkdir -p $tdir 2>/dev/null
    cp $UPARM/rvoemsao.par  $UPARM/imhimcome.par $UPARM/rvocontps.par $tdir
    export UPARM=$tdir
#
# Get sky fiber list from sky file names
#
    cat $skylist   |
    while read skyfiblist; do
      skynum=$((skynum + 1))
      nskyfibers=$(echo $skyfiblist | tr "+" "\012" | wc -l)
      skyfibers="$(    \
       (for file in $(echo $skyfiblist | tr '+' ' '); do
          table -h <${prefix}keyval.db 'HDRFILE == "'$file'" {print FIBER}'
        done) | tr "\012" "_")"

      extn=${skyfiblist##*\.}
      extn=${extn%%+}
      skysumfile=${tmp}${skyfibers}.${procpid}.${extn}
      shortname=${tmp}skyset_${seqno}_${skynum}.$extn

      /bin/rm -f $skysumfile $shortname
      echo $skyfiblist | tr '+' '\n' >${tdir}skyfiles

      callimages imcombine @${tdir}skyfiles $shortname 'rejmask=""'  \
        'combine="average"' 'reject="avsigclip"'  'outtype="real"'   \
        'scale="none"'  'zero="none"' mclip=yes  lsigma=3. hsigma=3. \
        sigscale=0.1  'logfile=""' project=no
      if((skyfit == 0)); then
        callrvsao emsao $shortname report=-4 mode=q mincont=10000. |  
        headoff lht01 leqw01                             |
        read skyEmiss skyArea
        xit=$?
        [[ $xit != 0 ]] && {
          print "$(basename $0): Error=$xit in emsao, seq=$seqno \?\n"
        }
      else
        skyEmiss=0.0
        skyArea=0.0
      fi
      printf "%s\t%s\t%s\t%s\t%s\t%s\t%s\n" $skyfiblist $skysumfile $shortname\
         "$nskyfibers" $skyfibers $skyEmiss $skyArea >>${skylist}.db
    done
  ) &
done

wait

cat ${tmp}skyflist*.db |
fldtotable Skyfilelist Skysumfile Skyabbrev Nskyfibers Skyfibers \
    skyEmiss skyArea  |
sorttable Skyfilelist >$tmpfileA

#cp $tmpfileA ./fileA.skylist.db

/bin/rm -rf ${tmp}skyfilelist ${tmp}skyflist*

#cp ${prefix}objlist.db /tmp/objlist.db
#cp $tmpfileA /tmp/fileA
  
#printf \
# "\n skyproc %s: input skyweight=%s\n\n" $(basename $(pwd)) "$skyweight" 1>&2

sorttable Skyfilelist <${prefix}objlist.db              |
jointable -j Skyfilelist  -  $tmpfileA                  |
column -a ratio lineratio arearatio                     |
if((skyfit == 0)); then
  compute '
     lineratio=sprintf("%.6f", objEmiss/skyEmiss)
     arearatio=sprintf("%.6f", objArea/skyArea)
     if ('$skyweight' < 0.01) {
        if('$usearea' == 0) { ratio = lineratio } else { ratio = arearatio }
     } else {
        ratio = '$skyweight'
     }' 
else
  compute ' lineratio=0.0; arearatio=0.0; ratio=0.0'  >$tmpfileB
fi >$tmpfileB

sorttable HDRFILE <$tmpfileB >${prefix}objlist.db

if ((verbose > 0)); then
   printf "\nElapsed time after summed skys: %.2f seconds\n" \
      $SECONDS
   times
fi

# Make sure output directory is there
[[ -z "$destdir" || -d $destdir ]] || mkdir -p $destdir

seqno=0

((verbose > 1)) && printf "\n"

columns="HDRFILE Skysumfile Skyabbrev Nskyfibers \
         Skyfibers Apertures objEmiss skyEmiss ratio"

nxtjob=0
waitjob=0

headoff $columns <${prefix}objlist.db |
while read $columns; do
  seqno=$((seqno + 1))

# Fork off up to $mpjobs subshells into the background, waiting for
# one before doing the next

 (
  temp=${tmp}$seqno/
  trap \
   'print -u2 "B:ERROR $? in $0 seq $seqno at line $LINENO"; exit 100' ERR
  trap '/bin/rm -rf $temp' EXIT

  mkdir -p $temp 2>/dev/null

  if ((verbose > 1)); then
    printf "%s: Obj %s height= %.1f, Sky %s= %.1f, ratio= %.3f\n" \
           $HDRFILE "$emline" $objEmiss "$emline" $skyEmiss $ratio
    printf "%s: Obj %s area  = %.1f, Sky area= %.1f, ratio= %.3f\n" \
           $HDRFILE "$emline" $objArea $skyArea $ratio
  fi

  extn=$(expr $HDRFILE : '.*\(\..*\)$' \| $HDRFILE)
  outbase=$(basename $HDRFILE $extn)
  outfile=${temp}${outbase}.sub$extn
  destfile=${destdir}${outbase}.ms${extn}
  tmpdestfile=${temp}${outbase}.ms${extn}  

  #
  # for multiprocessing, need to make separate copy of the parameters file
  # so they do not conflict with each other
  
  cp ${UPARM}imlimexpr.par  ${UPARM}imlimstak.par \
     ${UPARM}imlimrene.par  ${UPARM}imlimdele.par  ${temp}

  UPARM=$temp
  export UPARM  

  if((skyfit == 0)); then
    callimages imexpr '"a - b * c"' $outfile $HDRFILE $Skyabbrev \
         $ratio verbose=no mode=q  1>/dev/null
    exitval=$?

  ((exitval != 0)) && print -u2 -f "ERROR: imexpr: status $exitval\n" 

   # get N by 1 files
    callimages imstack $outfile ${temp}obj$extn
    callimages imstack $HDRFILE ${temp}raw$extn
    callimages imstack $Skyabbrev ${temp}sky$extn

   # now combine into one N by 1 by 3 file
    callimages imstack \
       "${temp}obj${extn},${temp}raw${extn},${temp}sky${extn}" $tmpdestfile 
    exitval=$?

    callimages imdelete $outfile verify=no; xit=$?
    ((xit != 0)) && {
      print "$(basename $0): can't imdelete ${outfile}\?\n";  exit $xit
    }

  else
    if((verbose > 0)); then vb="verbose=yes"; else vb="verbose=no"; fi
    callhectospec hskyfit1 $HDRFILE $Skyabbrev $tmpdestfile $vb debug=no
  fi

  
  delhead BANDID1 BANDID2 BANDID3 BANDID4 $tmpdestfile
  now=$(getdate 2fd now)
  datemsg=$(printf "Sky subtracted %s" $now)
  if ((skyweight > 0.001)); then 
      skyscale=$(printf "FixedScale=%.3f" $skyweight)
  elif ((usearea == 0)); then  
      skyscale="LineRatio_${emline}"
  else
      skyscale="AreaRatio_${emline}"
  fi
  sethead BANDID1='spectrum - scaled sky'                        \
          BANDID2='raw spectrum'                                 \
          BANDID3='averaged sky'                                 \
          SKYPROC="$datemsg Rev $REV"                            \
          NSKYFIB=$Nskyfibers                                    \
          SKYFIB="$Skyfibers"                                    \
          SKYAPER="$Apertures"                                   \
          SKYCERT=$skycert                                       \
          SKYSTDLM=$skystdlim                                    \
          NSKYMAX=$nskys                                         \
          SKYMETHD="$skymethod"                                  \
          $tmpdestfile 
  if((skyfit == 0)); then
    sethead SKYDIST=$fiberdist                                   \
            SKYRATIO=$ratio                                      \
            SKYSCALE="$skyscale"                                 \
            $tmpdestfile 
  fi
 # Fix for when APNUM1 keyword not passed through
  apname=$(gethead APNUM1 $tmpdestfile)
  if [[ -z $apname ]]; then
    aperture=$(gethead APERTURE $tmpdestfile)
    [[ -n $aperture ]] && {
      apnum=$(printf "%s %s" $aperture $aperture)
      sethead $tmpdestfile APNUM1="$apnum"
    }
  fi

  mv $tmpdestfile $destfile
  
  ((verbose > 1)) &&  printf "Output: %s\n\n" $destfile

  exit $exitval
 )  >${tmp}logfile.$seqno 2>&1  &  # END FORKED SUBSHELL

  ((verbose > 1)) && printf " Forked job %d for %s\n" $seqno $HDRFILE
  bkgnd[$nxtjob]=$!
  hdrfile[$nxtjob]=$HDRFILE
  nxtjob=$((nxtjob + 1)) 

# if mpjobs jobs are not running, continue forking before waiting
  ((nxtjob - waitjob < mpjobs)) && continue
  
  wait ${bkgnd[$waitjob]}
  retval=$?

  retarr[$waitjob]=$retval
  ((retval != 0)) && {
     printf "\nERROR val %d while skyproc.sub of %s\n" \
             $retval ${hdrfile[$waitjob]}
    exitval=$retval
  }
  waitjob=$(($waitjob + 1))
  cat ${tmp}logfile.$waitjob
  /bin/rm -f ${tmp}logfile.$waitjob
done

# wait for last several to finish
while ((waitjob < nxtjob)) do
  wait ${bkgnd[$waitjob]}
  retval=$?
#  printf "Wait: %d %d ret = %d\n" $waitjob ${bkgnd[$waitjob]} $retval 
  retarr[$waitjob]=$retval
  ((retval != 0)) && {
     printf "\nERROR val %d while skyproc.sub of %s\n"   \
          $retval ${hdrfile[$waitjob]}
     exitval=$retval
  }
  waitjob=$((waitjob + 1))
  cat ${tmp}logfile.$waitjob
done  

((verbose > 0)) && {
   printf "\nElapsed time at exit: $SECONDS seconds\n"
   times
}

[[ -e $destdir/rvoemsao.par ]] && 
   mv $destdir/rvoemsao.par $destdir/hecskemso.par
[[ -e $destdir/rvoxcsao.par ]] && 
   mv $destdir/rvoxcsao.par $destdir/hecskxcso.par
[[ -e $destdir/rvocontps.par ]] && 
   mv $destdir/rvocontps.par $destdir/hecskcons.par

exit $exitval

# Aug 26 2004	DJM Process SKY spectra like OBJECT spectra
# Aug 26 2004	DJM Add SKYPROC keyword (will add test and drop out later)
# Oct 14 2004	DJM Add test for no sky files with error message and exit
# Nov 17 2004   WFW Add radius > 306 to reject parked fibers
# Mar  6 2006   WFW Add SKYFIT option, using hskyfit1 and closest-by-aperture
#                  skys.
